-- Turbine Imports..
import "Turbine";
import "Turbine.Gameplay";
import "Turbine.UI";
import "Turbine.UI.Lotro";

-- Plugin Imports..
import "CubePlugins.DeedTracker.Timer";
import "CubePlugins.DeedTracker.Globals";
import "CubePlugins.DeedTracker.Strings";
import "CubePlugins.DeedTracker.PluginFunctions";
import "CubePlugins.DeedTracker.ChatLogger";
import "CubePlugins.DeedTracker.Commands";
import "CubePlugins.DeedTracker.GeneralFunctions";
import "CubePlugins.DeedTracker.Onscreen";
import "CubePlugins.DeedTracker.VindarPatch";
import "CubePlugins.DeedTracker.DropDown";
import "CubePlugins.DeedTracker.Images";
import "CubePlugins.DeedTracker.DeedIDsToIndicies";
import "CubePlugins.DeedTracker.RegionInfo";
import "CubePlugins.DeedTracker.ConflictingQuests";
import "CubePlugins.DeedTracker.DeedItems";

-- Language Specific..
if (LANGUAGE == "DE") then
    import "CubePlugins.DeedTracker.DeedInfoDE";
    import "CubePlugins.DeedTracker.DeedNamesToIDsDE";
    import "CubePlugins.DeedTracker.DeedObjectiveInfoDE";
elseif (LANGUAGE == "FR") then
    import "CubePlugins.DeedTracker.DeedInfoFR";
    import "CubePlugins.DeedTracker.DeedNamesToIDsFR";
    import "CubePlugins.DeedTracker.DeedObjectiveInfoFR";
else
    import "CubePlugins.DeedTracker.DeedInfoEN";
    import "CubePlugins.DeedTracker.DeedNamesToIDsEN";
    import "CubePlugins.DeedTracker.DeedObjectiveInfoEN";
end

-- Windows --
import "CubePlugins.DeedTracker.Options";
import "CubePlugins.DeedTracker.MainWin";
import "CubePlugins.DeedTracker.DeedTooltipWindow";
import "CubePlugins.DeedTracker.StringTooltipWindow";
import "CubePlugins.DeedTracker.DisambiguateWindow";
import "CubePlugins.DeedTracker.DeedOrQuestWindow";
import "CubePlugins.DeedTracker.CompletionWindow";
import "CubePlugins.DeedTracker.CompanionImportWindow";
import "CubePlugins.DeedTracker.MinimizedIcon";

function UnloadReloader()
    Turbine.PluginManager.RefreshAvailablePlugins();
    local loadedPlugins = Turbine.PluginManager.GetLoadedPlugins();

    reloadedUnloader = Turbine.UI.Control();
    reloadedUnloader:SetWantsUpdates( true );
    
    reloadedUnloader.Update = function( sender, args )
        Turbine.PluginManager.RefreshAvailablePlugins();
        local loadedPlugins = Turbine.PluginManager.GetLoadedPlugins();

        for k,v in pairs(loadedPlugins) do
            if v.Name == "~Deed Tracker Reloader" then
                Turbine.PluginManager.UnloadScriptState( 'DeedTrackerReloader' );
            end
        end
        reloadedUnloader:SetWantsUpdates( false );
    end
end

function GetCharDataFilename(characterName)
    local fixedName = characterName;

    if (fixedName:sub(1,1) == "~") then
        fixedName = fixedName:sub(2);
    end

    fixedName = fixedName:gsub("-", "_");
    return "DeedTracker_CharData_" .. fixedName;
end

-- v1 format: All characters in one file, no separate :CHARACTERS entry in _CHARDATA
-- v2 format: All characters in separate files, :CHARACTERS entry in _CHARDATA tells us which files to look for
function saveData()
    saveCharacterSettings();
    saveCurrentCharacterData();
    saveServerSettings();

    Info("  Deed Tracker: Save complete");
end

function saveCharacterSettings()
    PatchDataSave(Turbine.DataScope.Character, "DeedTracker_Settings", SETTINGS);
end

function saveServerSettings()
    local serverData = {};
    -- Write out special categories specially:
    for k,v in pairs (_CHARDATA) do
        if (k:sub(1, 1) == ":" and k ~= MYCHAR:GetName()) then -- Don't do special categories or the current charcater
            serverData[k] = deepcopy(v);
        end
    end
    PatchDataSave(Turbine.DataScope.Server, "DeedTracker_CharData", serverData);
end

function saveCurrentCharacterData()
    if (IsSessionPlay()) then
        return;
    end

    local charName = MYCHAR:GetName();
    Debug("Deed Tracker: Saving character data for " .. charName .. "...");

    local mainWin = DeedTrackerWin.GetInstance();
    mainWin:SaveRegionExpandedStates();

    CompletionWindow.GetInstance():SaveUnclearedDeeds();

    local filename = GetCharDataFilename(charName);
    PatchDataSave(Turbine.DataScope.Server, filename, _CHARDATA[charName]);
end

function loadCharData(charName)
    local filename = GetCharDataFilename(charName);
    local savedCharData = PatchDataLoad(Turbine.DataScope.Server, filename);
    if (type(savedCharData) == 'table') then
        _CHARDATA[charName] = deepcopy(savedCharData);
    end
end

function loadData()
    local SavedSettings = {};
    SavedSettings = PatchDataLoad(Turbine.DataScope.Character, "DeedTracker_Settings");

    -- Check the saved settings to make sure it is still compatible with newer updates, add in any missing default settings
    if type(SavedSettings) == 'table' then
        local tempSETTINGS = {};
        tempSETTINGS = deepcopy(DEFAULT_SETTINGS);
        SETTINGS = mergeTables(tempSETTINGS,SavedSettings);
    else
        SETTINGS = deepcopy(DEFAULT_SETTINGS);
    end

    -- Load Server Data:
    local SavedCharData = {};
    SavedCharData = PatchDataLoad(Turbine.DataScope.Server, "DeedTracker_CharData");

    local isSaveFilePresent = false;
    -- Check the saved settings to make sure it is still compatible with newer updates, add in any missing default settings
    if type(SavedCharData) == 'table' then
        _CHARDATA = deepcopy(SavedCharData);
        isSaveFilePresent = true;
    end

    -- If this is V1 (all-in-one), update to V2 (per-character save file):
    if (_CHARDATA[":CHARACTERS"] == nil) then
        if (SETTINGS.VERBOSE_OUTPUT) then
            Debug("Deed Tracker: Upgrading save file format from v1 to v2");
        end

        -- Save out a backup in the original format if this isn't a brand new install:
        if (isSaveFilePresent) then
            Debug("Deed Tracker: Backing up save file to 'DeedTracker_CharData_v1backup'...");
            PatchDataSave(Turbine.DataScope.Server, "DeedTracker_CharData_v1backup", _CHARDATA);
            Debug("  Deed Tracker: save file backup complete.");
        end  

        -- Mark the version:
        _CHARDATA[":CHARDATA_VERSION"] = "v2.0";

        -- Store the character names:
        _CHARDATA[":CHARACTERS"] = GetAltNames();

        -- Write out a file for every existing character:
        for k,v in pairs (_CHARDATA) do
            if (k:sub(1, 1) ~= ":") then -- Don't do special categories
                Debug("Deed Tracker: Creating individual save file for " .. k .. "...");

                local filename = GetCharDataFilename(k);
                PatchDataSave(Turbine.DataScope.Server, filename, _CHARDATA[k]);

                Debug("  Deed Tracker: individual save file created.");
            end
        end

        -- Remove all char data from core file:
        saveServerSettings();

        -- v1 has been converted to v2, continue.
    else -- Already V2
        -- Load character data for all known characters:
        -- V1: This is already included in the above DeedTracker_CharData file
        -- V2: Iterate through each character in :CHARACTERS, load their file.
        for i=1, table.maxn(_CHARDATA[":CHARACTERS"]) do
            if (_CHARDATA[":CHARACTERS"][i] == nil) then
                table.remove(_CHARDATA[":CHARACTERS"], i);
            else
                local name = _CHARDATA[":CHARACTERS"][i];
                loadCharData(name);
            end
        end
    end

end

function checkForCompanionImport()
    -- Check to see if a LOTRO Companion import file is present:
    local companionImport = {};
    companionImport = PatchDataLoad(Turbine.DataScope.Character, "DeedTracker_CompanionImport");
    if type(companionImport) == 'table' and companionImport["SKIP"] == nil then
        processCompanionImport(companionImport);

        -- Now that the import is done, clear the file so we don't import it again:
        local emptyCompanion = {};
        emptyCompanion["SKIP"] = true;
        PatchDataSave(Turbine.DataScope.Character, "DeedTracker_CompanionImport", emptyCompanion);
    end

end

function processCompanionImport(companionImport)
    local charName = MYCHAR:GetName();
    local companionImportWindow = CompanionImportWindow.GetInstance();

    for i=1, table.maxn(_DEED_LOG_PAGES) do
        for j=1, table.maxn(_DEED_LOG_PAGE_TABS[i]) do
            for k=1, table.maxn(_DEED_LOG_PAGE_TAB_CONTENTS[i][j]) do
                local currentDeed = _DEED_LOG_PAGE_TAB_CONTENTS[i][j][k];
                -- Get the save index for this deed:
                -- skip categories:
                if (not IsCategory(currentDeed)) then
                    local id = currentDeed.ID;

                    local companionComplete = companionImport[id] ~= nil and companionImport[id] == "COMPLETED";
                    local deedComplete = GetDeedComplete(charName, i, j, k);

                    if (companionComplete and not deedComplete) then
                        -- If Companion thinks it is complete and Deed Tracker does not:
                        --   Then set it to true and let the player know
                        companionImportWindow:AddDeedToBeCompleted(i, j, k);
                    elseif (not companionComplete and deedComplete) then
                        -- If Companion thinks it is not complete and Deed Tracker does:
                        --   Then let the player know about the discrepancy        
                        companionImportWindow:AddCompletedDeedThatIsNotDone(i, j, k);
                    end

                end
            end
        end
    end
    companionImportWindow:Show();

end

function updateData()
    -- v1.1.0:
    -- Realized we shouldn't store possible character names in _CHARDATA, so changed "SEARCH" and "SERVER" to ":SEARCH" and ":SERVER".
    -- Convert existing ones, to make like easier for testers, but remove eventually in case someone *does* have these names.
    if (_CHARDATA["SEARCH"] ~= nil) then
        _CHARDATA[":SEARCH"] = _CHARDATA["SEARCH"];
        _CHARDATA["SEARCH"] = nil;
    end

    if (_CHARDATA["SERVER"] ~= nil) then
        if (_CHARDATA["SERVER"]["DEEDS"] ~= nil) then
            _CHARDATA["SERVER"]["DEEDS"] = nil;
        end
        _CHARDATA[":SERVER"] = _CHARDATA["SERVER"];
        _CHARDATA["SERVER"] = nil;
    end

end

function DrawWindows()
    DeedTrackerWin.GetInstance():DrawMainWin();
    CompletionWindow.GetInstance():LoadUnclearedDeeds();
end

Turbine.Plugin.Unload = function (sender, args)
    -- Save the data when the plugin unloads.
    saveData();
    Info(GetString(_LANG.STATUS.UNLOADED));
end

function SetServerLevelCap()
    local isLegendaryServer = LoadServerField("LEGENDARY_SERVER");

    local serverLevelCap = tonumber(LoadServerField("SERVER_LEVEL_CAP"));
    if (DEFAULT_SERVER_LEVEL_CAP > serverLevelCap) then
        Debug(string.format("Deed Tracker: Default Server Level (%d) is higher than stored server level (%d), using it instead", DEFAULT_SERVER_LEVEL_CAP, serverLevelCap));

        serverLevelCap = DEFAULT_SERVER_LEVEL_CAP;
        SaveServerField("SERVER_LEVEL_CAP", DEFAULT_SERVER_LEVEL_CAP);
    end

    if (isLegendaryServer) then
        CURRENT_LEVEL_CAP = tonumber(LoadServerField("LEGENDARY_SERVER_LEVEL_CAP"));
    else
        CURRENT_LEVEL_CAP = serverLevelCap;
    end
    Debug("Deed Tracker: Setting server level cap to " .. CURRENT_LEVEL_CAP);
end

function RegisterForCombatChanged()
    MYCHAR.InCombatChanged = function() 
        CompletionWindow.GetInstance():InCombatChanged();
        DisambiguateDialog.GetInstance():InCombatChanged();
        DeedTrackerWin.GetInstance():InCombatChanged();
    end
end

function DetermineIfWaypointIsAvailable()
    -- Determine if Waypoint is available:
    local loadedPlugins = Turbine.PluginManager:GetAvailablePlugins();
    for i=1, table.maxn(loadedPlugins) do
        if (loadedPlugins[i].Name == "Waypoint") then
            WAYPOINT_AVAILABLE = true;
        end
    end

end

function Main()
    -- Initiate load sequence
    UnloadReloader();
    loadData();
    updateData();
    CheckCharData();
    checkForCompanionImport();
    SetServerLevelCap();
    RegisterForCombatChanged();
    DrawWindows();
    if (IsSessionPlay()) then
        Debug(GetString(_LANG.STATUS.SESSION_PLAY));
    else
        InitiateChatLogger();
    end
    RegisterCommands();
    CreateOptionsContent();
    DetermineIfWaypointIsAvailable();

    -- Load text
    Info(GetString(_LANG.STATUS.LOADED));

    -- Check if session play completed a deed by delivering an item:
    CheckForDeedItems();
end

Main();
