
function GetString(array, language)
    if (language == nil or 
        language == "") then
        language = LANGUAGE;
    end
    if (array == nil) then
        return "";
    end
    if (language == "ENGLISH") then
        Debug("Changing language from ENGLISH to EN");
        language = "EN";
    end
    if (array[language] ~= nil) then
        return array[language];
    end
    return array["EN"];
end

function DebugFoundDeeds(foundDeeds)
    for deedIndex=1, table.maxn(foundDeeds) do
        local deed = foundDeeds[deedIndex];
        local i = deed["i"];
        local j = deed["j"];
        local k = deed["k"];
        local message = "  " .. GetFullyQualifiedDeedName(i,j,k);
        if (foundDeeds[deedIndex]["region_ok"]) then
            message = message .. " (same or adjacent region)";
        end
        Debug(message);
    end
end

function IfDeedMarkComplete(character, deedName)
    local mainWin = DeedTrackerWin.GetInstance();

    -- Check if the thing that was completed matches one of the deeds we know about.
    --Debug("Checking \"" .. deedName .. "\" to see if it is a deed.");

    local foundDeeds = GetDeedIndicies(character, deedName);
    local foundDeedCount = table.getn(foundDeeds);

    local regionOkCount = GetRegionOkCount(foundDeeds);

    if (foundDeedCount == 0) then
        Debug("Deed Tracker: Did not find an uncompleted deed matching \"" .. deedName .. "\"");
    elseif (foundDeedCount == 1) then
        local deed = foundDeeds[1];
        local i = deed["i"];
        local j = deed["j"];
        local k = deed["k"];

        local saveIndex = GetDeedSaveIndex(i, j, k);

        -- Note: Identical logic existis in DisambiguateWindow, extract out somewhere.
        -- Check if this is one of the deeds that has an identical name to a quest.
        -- If so, open the Deed or Quest dialog for confirmation.
        if (DoesDeedHaveConflictingQuest(i, j, saveIndex)) then
            if (WasThereAnnouncementOfConflictingQuest(i, j, saveIndex)) then
                Debug("Deed Tracker: Completed something that is both a deed and a quest, but the quest was announced in chat.");
            else
                Debug("Deed Tracker: Completed something that is both a deed and a quest. Asking user to choose.");
                local deedOrQuestWindow = DeedOrQuestDialog.GetInstance(deedName);
                deedOrQuestWindow:SetDeed(deed);
            end
        -- Otherwise, mark the deed complete
        else
            IS_DEED_DETECTED = true;
            mainWin:MarkDeedCompleteFromChat(i, j, k);
            IS_DEED_DETECTED = false;
        end
    elseif (regionOkCount == 1) then
        Debug("Found the following matching deeds, but only 1 is in an adjacent region:");
        DebugFoundDeeds(foundDeeds);

        local onlyGoodRegionIndex = GetFirstRegionOkIndex(foundDeeds);
        local deed = foundDeeds[onlyGoodRegionIndex];
        local i = deed["i"];
        local j = deed["j"];
        local k = deed["k"];
        IS_DEED_DETECTED = true;
        mainWin:MarkDeedCompleteFromChat(i, j, k);
        IS_DEED_DETECTED = false;
    elseif (foundDeedCount == 2 and 
        AreDeedsPaired(foundDeeds)) then -- "The Mines of Moria" and "The Mines of Moria"
        for deedIndex=1, table.maxn(foundDeeds) do
            local i = foundDeeds[deedIndex]["i"];
            local j = foundDeeds[deedIndex]["j"];
            local k = foundDeeds[deedIndex]["k"];
            IS_DEED_DETECTED = true;
            mainWin:MarkDeedCompleteFromChat(i, j, k);
            IS_DEED_DETECTED = false; 
        end
    else
        Debug("Found the following matching deeds (" .. regionOkCount .. " in an adjacent region), asking player to choose:");
        DebugFoundDeeds(foundDeeds);

        local window = DisambiguateDialog.GetInstance();
        window:SetDeeds(deedName, foundDeeds);
    end
end

function IsValueNotEmpty(value)
    local result =
        value ~= nil and
        value ~= "" and
        value ~= 0 and
        value ~= "0";
    return result;
end

function GetCrv(crv)
    if (IsValueNotEmpty(crv)) then
        return crv;
    else
        return "";
    end
end

function DoesDeedHaveSubtype(subtype)
    return IsValueNotEmpty(subtype);
end

function IsDeedNotAvailable(deedNa)
    return IsValueNotEmpty(deedNa);
end

function GetStringSubtype(deedCrv, deedSubtype)
    local result = "unknown subtype";
    if (deedCrv == "Race") then
        if (_RACES[deedSubtype] == nil) then
            result = "unknown race";
        else
            result = GetString(_RACES[deedSubtype]);
        end
    elseif (deedCrv == "Class") then
        if (_CLASSES[deedSubtype] == nil) then
            result = "unknown class";
        else
            result = GetString(_CLASSES[deedSubtype]);
        end
    elseif (deedCrv == "Vocation") then
        if (_VOCATIONS[deedSubtype] == nil) then
            result = "unknown vocation";
        else
            result = GetString(_VOCATIONS[deedSubtype]);
        end
    end
    return result;
end

function GetStringNotAvailable(deed)
    local result = "";

    if (IsDeedNotAvailable(deed.NA)) then
        result = GetString(_NOT_ACTIVE_REASON[deed.NA]);
    end

    return result;
end

function GetFullyQualifiedDeedName(i, j, k)
    local deedName = "[[Warning: Could not find deed " .. i .. ", " .. j .. ", " .. k .. "]]";
    local specifier = "";
    local deed = _DEED_LOG_PAGE_TAB_CONTENTS[i][j][k];
    if (deed ~= nil) then
        local deedCrv = GetCrv(deed.CRV);
        local subtype = deed.SUBTYPE;
        if (DoesDeedHaveSubtype(subtype)) then
            specifier = " (" .. GetStringSubtype(deedCrv, subtype) .. ")";
        end

        if (IsDeedNotAvailable(deed.NA)) then
            specifier = " (" .. GetStringNotAvailable(deed) .. ")";
        end

        local note = GetString(deed.NOTE);

        if (note ~= "") then
            specifier = specifier .. " (" .. note .. ")";
        end

        deedName = 
            GetString(_DEED_LOG_PAGES[i]) .. " / " .. 
            GetString(_DEED_LOG_PAGE_TABS[i][j]) .. " / " .. 
            GetString(_DEED_LOG_PAGE_TAB_CONTENTS[i][j][k].NAME) ..
            specifier;
    end

    return deedName;
end

function GetClass(character)
    local class = 0;
    if (_CHARDATA[character]["CLASS"] ~= nil) then
        class = _CHARDATA[character]["CLASS"];
    end
    return class;
end

function GetVocation(character)
    local vocation = 0;
    if (_CHARDATA[character]["VOCATION"] ~= nil) then
        vocation = _CHARDATA[character]["VOCATION"];
    end
    return vocation;
end

function GetRace(character)
    local race = 0;
    if (_CHARDATA[character]["RACE"] ~= nil) then
        race = _CHARDATA[character]["RACE"];
    end
    return race;
end

function GetLevel(character)
    local level = 0;
    if (_CHARDATA[character].LEVEL ~= nil) then
        level = tonumber(_CHARDATA[character].LEVEL);
    end
    return level;
end

function GetDeedSaveIndex(i, j, k)
    local saveIndex = -1;
    if (_DEED_LOG_PAGE_TAB_CONTENTS[i] == nil) then
        Debug("Error with i: GetDeedSaveIndex(" .. i .. ", " .. j .. ", " .. k .. ")");
    end
    if (_DEED_LOG_PAGE_TAB_CONTENTS[i][j] == nil) then
        Debug("Error with j: GetDeedSaveIndex(" .. i .. ", " .. j .. ", " .. k .. ")");
    end
    if (_DEED_LOG_PAGE_TAB_CONTENTS[i][j][k] == nil) then
        Debug("Error with k: GetDeedSaveIndex(" .. i .. ", " .. j .. ", " .. k .. ")");
    end

    if (_DEED_LOG_PAGE_TAB_CONTENTS[i][j][k]["SAVE_INDEX"] ~= nil) then
        saveIndex = _DEED_LOG_PAGE_TAB_CONTENTS[i][j][k]["SAVE_INDEX"];
    end
    return saveIndex;
end

function GetCurrentKFromSaveIndex(i, j, saveIndex)
    for k=1, table.maxn(_DEED_LOG_PAGE_TAB_CONTENTS[i][j]) do
        if (_DEED_LOG_PAGE_TAB_CONTENTS[i][j][k]["SAVE_INDEX"] == saveIndex) then
            return k;
        end
    end
    return -1;
end

function SetDeedComplete(character, i, j, k, isDeedComplete)
    local mainWin = DeedTrackerWin.GetInstance();

    if (isDeedComplete) then
        -- If someone is checking off deeds, they don't need the chat spam.
        if (SETTINGS.VERBOSE_OUTPUT or IS_DEED_DETECTED) then
            Info(GetString(_LANG.DEEDS.COMPLETED) .. " " .. GetFullyQualifiedDeedName(i,j,k) .. "!");
        end
    else
        Debug(GetString(_LANG.DEEDS.UNCOMPLETED) .. " " .. GetFullyQualifiedDeedName(i,j,k));
    end
    local saveIndex = GetDeedSaveIndex(i, j, k);
    local saveEntry = MakeSaveEntry(isDeedComplete);
    local isNotAvailable = IsDeedNotAvailable(_DEED_LOG_PAGE_TAB_CONTENTS[i][j][k].NA);
    local change = -1;
    if (isDeedComplete) then
        change = 1;
    end
    if (isNotAvailable) then
        change = 0;
    end

    if (saveIndex ~= -1) then
        _CHARDATA[character]["DEEDS"][i][j][saveIndex] = saveEntry;
        _CHARDATA[character]["DEEDS"]["COMPLETED"] = _CHARDATA[character]["DEEDS"]["COMPLETED"] + change;

        -- (Re-)Start 5 second timer, then Save data.
        SaveDeedTimer:Start();

        -- if the Disambiguation window is open for this deed, close it:
        local deedName = GetString(_DEED_LOG_PAGE_TAB_CONTENTS[i][j][k].NAME);
        local window = DisambiguateDialog.GetInstance();
        window:ClearAmbigiousDeed(deedName);
    end

    -- If completing a deed, should we check all deeds below it?
    -- If un-completing a deed, should we un-check all deeds above and below it?
    if (SETTINGS.CASCADE_COMPLETION and not IS_IMPORT_HAPPENING) then
        local deedsInTab = table.getn(_DEED_LOG_PAGE_TAB_CONTENTS[i][j]);
        local deedTier = GetDeedTier(i, j, k);
        local correctLevel = true;
        local parentLevel = deedTier - 1;

        if (isDeedComplete or IS_DEED_SHIFT_CLICKED) then
            for deedIndex=k+1, deedsInTab do
                local hideDeed, skipDeed, lvlTooHigh, isCategory, isNotAvailable = GetDeedSkipInfo(character, _DEED_LOG_PAGE_TAB_CONTENTS[i][j][deedIndex], i, j, deedIndex);
                if (not skipDeed and not isCategory) then
                    local tempDeedTier = GetDeedTier(i, j, deedIndex);
                    local isAlreadyComplete = GetDeedComplete(character, i, j, deedIndex);
                    if (correctLevel and tempDeedTier == (deedTier + 1) and isDeedComplete ~= isAlreadyComplete) then
                        mainWin:MarkDeedCompleteFromChat(i, j, deedIndex);
                    elseif (tempDeedTier == deedTier) then
                        correctLevel = false;
                    end
                end
            end
        elseif (not ARE_WE_CASCADE_DESELECTING) then
            ARE_WE_CASCADE_DESELECTING = true;
            -- Un-completing a deed meens any completed parent deed should also be unchecked.
            -- Look for level - 1, level - 2, etc. until we find level == 0 or k is 1.
            for deedIndex=k-1, 1, -1 do
                local tempDeedTier = GetDeedTier(i, j, deedIndex);
                local isAlreadyComplete = GetDeedComplete(character, i, j, deedIndex);
                if (correctLevel and tempDeedTier == parentLevel) then
                    if (isAlreadyComplete) then
                        mainWin:MarkDeedCompleteFromChat(i, j, deedIndex);
                    end
                    parentLevel = parentLevel - 1;
                elseif (parentLevel < 0) then
                    correctLevel = false;
                end

            end
            ARE_WE_CASCADE_DESELECTING = false;
        end
    end

end

function GetDeedTier(i, j, k)
    local deedTier = 0;
    if (_DEED_LOG_PAGE_TAB_CONTENTS[i][j][k].TIER ~= nil) then
        deedTier = _DEED_LOG_PAGE_TAB_CONTENTS[i][j][k].TIER;
    end
    return deedTier;
end

function GetDeedComplete(character, i, j, k)
    local saveIndex = GetDeedSaveIndex(i, j, k);
    if (saveIndex == -1) then
        return false;
    end

    -- Temporary function to deel with any of my characters with the old bool-only save system.
    if (type(_CHARDATA[character]["DEEDS"][i][j][saveIndex]) == "boolean") then
        _CHARDATA[character]["DEEDS"][i][j][saveIndex] = MakeSaveEntry(_CHARDATA[character]["DEEDS"][i][j][saveIndex]);
    end
    -- 

    local saveValue = _CHARDATA[character]["DEEDS"][i][j][saveIndex];
    if (saveValue == nil) then
        return false;
    end
    return saveValue.COMPLETE;
end

function MarkConflictingQuestComplete(i, j, saveIndex)
    local character = MYCHAR:GetName();
    local saveEntry = MakeSaveEntry(true);

    if _CHARDATA[character]["QUESTS"] == nil then _CHARDATA[character]["QUESTS"] = {} end;
    if _CHARDATA[character]["QUESTS"][i] == nil then _CHARDATA[character]["QUESTS"][i] = {} end;
    if _CHARDATA[character]["QUESTS"][i][j] == nil then _CHARDATA[character]["QUESTS"][i][j] = {} end;
    
    _CHARDATA[character]["QUESTS"][i][j][saveIndex] = saveEntry;
end

function GetConflictingQuestComplete(i, j, saveIndex)
    local character = MYCHAR:GetName();

    local questCompleted =
        _CHARDATA[character]["QUESTS"] ~= nil and
        _CHARDATA[character]["QUESTS"][i] ~= nil and
        _CHARDATA[character]["QUESTS"][i][j] ~= nil and
        _CHARDATA[character]["QUESTS"][i][j][saveIndex] ~= nil and
        _CHARDATA[character]["QUESTS"][i][j][saveIndex].COMPLETE;
    return questCompleted;
end

function GetRegionOkCount(foundDeeds)
    local count = 0;
    for deedIndex=1, table.maxn(foundDeeds) do
        if (foundDeeds[deedIndex]["region_ok"]) then
            count = count + 1;
        end
    end
    return count;
end

function GetFirstRegionOkIndex(foundDeeds)
    for deedIndex=1, table.maxn(foundDeeds) do
        if (foundDeeds[deedIndex]["region_ok"]) then
            return deedIndex;
        end
    end
    return 1;
end

function GetDeedMinimumLevel(deed)
    local minimumLevel = 0;
    if (deed.MIN_LVL ~= nil) then
        if (deed.MIN_LVL == "CAP") then
            return tonumber(CURRENT_LEVEL_CAP);
        else
            return tonumber(deed.MIN_LVL);
        end
    end
    return minimumLevel;
end

function GetDeedIndicies(character, deedName)
    local results = {};
    local resultCount = 0;

    local class = GetClass(character);
    local vocation = GetVocation(character);
    local race = GetRace(character);

    local deedNameTrimmed = remove_duplicate_spaces(trim(deedName));
    local deedNameTrimmerLower = deedNameTrimmed:lower();

    local matchingIDs = {};
    if (_NAMES_TO_IDS[deedNameTrimmed] ~= nil) then
        matchingIDs = _NAMES_TO_IDS[deedNameTrimmed];
    elseif(_NAMES_TO_IDS[deedNameTrimmerLower] ~= nil) then
        matchingIDs = _NAMES_TO_IDS[deedNameTrimmerLower];
    end

    for idIndex=1, table.getn(matchingIDs) do
        local indicies = _ID_LOOKUP[matchingIDs[idIndex]];
        local i = indicies.i;
        local j = indicies.j;
        local k = indicies.k;

        -- skip if already completed:
        local isAlreadyComplete = GetDeedComplete(character, i, j, k);
        local deed = _DEED_LOG_PAGE_TAB_CONTENTS[i][j][k];
        local checkboxExistsAndIsChecked = 
            deed["CHECK"] ~= nil and
            deed["CHECK"]:IsChecked();
        -- skip if category:
        local isCategory = IsCategory(deed);
        -- skip if we're too low level, as it's probably an instance quest with the same name.
        -- (e.g. Woe of the Willow -- Tier 1)
        local minLevel = GetDeedMinimumLevel(deed);
        local charLevel = MYCHAR:GetLevel();

        if (isAlreadyComplete or checkboxExistsAndIsChecked) then
            Debug("Skipping " .. GetFullyQualifiedDeedName(i,j,k) .. ", already completed");
        elseif (isCategory) then
            Debug("Skipping " .. GetFullyQualifiedDeedName(i,j,k) .. ", it's a category");
        elseif (minLevel > charLevel) then
            Debug("Skipping possible deed " .. GetFullyQualifiedDeedName(i,j,k) .. " because its min level (" .. minLevel .. ") is above character level (" .. charLevel .. ").");
        else
            local crv = GetCrv(deed.CRV);
            local isRace = false;
            local raceMatches = false;
            local isClass = false;
            local classMatches = false;
            local isVocation = false;
            local vocationMatches = false;
            local subtype = _DEED_LOG_PAGE_TAB_CONTENTS[i][j][k].SUBTYPE;
            if (DoesDeedHaveSubtype(subtype)) then
                if (crv == "Race") then -- Race
                    isRace = true;
                    raceMatches = race == subtype;
                elseif (crv == "Class") then -- Class
                    isClass = true;
                    classMatches = class == subtype;
                elseif (crv == "Vocation") then -- Vocation
                    isVocation = true;
                    vocationMatches = vocation == subtype;
                end
            end
            if (
                (not isRace or raceMatches) and
                (not isClass or classMatches)
                ) then
                local regionOK = false;
                -- If this is a regional deed, check if our region is ok.
                if (IsRegionDeed(i)) then
                    --Debug("Matching deed " .. GetFullyQualifiedDeedName(i,j,k) .. " is a regional deed");
                    local areRegionsAdjacent = AreRegionsAdjacent(LOCATION_NUMBER, i, j);
                    if (areRegionsAdjacent) then -- Our location is adjacent (or same as) the deed
                        --Debug("Matching deed " .. GetFullyQualifiedDeedName(i,j,k) .. " is adjacent to " .. GetRegionName(LOCATION_NUMBER));
                        regionOK = true;
                    else
                        --Debug("Matching deed " .. GetFullyQualifiedDeedName(i,j,k) .. " is not adjacent to " .. GetRegionName(LOCATION_NUMBER));
                        --regionOK = false;
                    end
                else
                    --Debug("Matching deed " .. GetFullyQualifiedDeedName(i,j,k) .. " is not a regional deed");
                end
                resultCount = resultCount + 1;
                results[resultCount] = { ["i"] = i; ["j"] = j; ["k"] = k; ["region_ok"] = regionOK; };
                --Debug("Found a matching deed: " .. GetFullyQualifiedDeedName(i,j,k));
                if (isVocation and not vocationMatches) then
                    local name = MYCHAR:GetName();
                    if (name == character) then
                        _CHARDATA[character]["VOCATION"] = MYCHAR:GetAttributes():GetVocation();
                        Debug(GetString(_LANG.DEEDS.UNKNOWN_VOCATION));
                        DeedTrackerWin.GetInstance():RefreshDeeds();
                    end
                end
            end
        end
    end

    return results;
end

function AreDeedsPaired(foundDeeds)
    local count = table.getn(foundDeeds);
    if (count ~= 2) then return false; end

    local i1 = foundDeeds[1]["i"];
    local j1 = foundDeeds[1]["j"];
    local k1 = foundDeeds[1]["k"];

    local i2 = foundDeeds[2]["i"];
    local j2 = foundDeeds[2]["j"];
    local k2 = foundDeeds[2]["k"];

    local deed1 = _DEED_LOG_PAGE_TAB_CONTENTS[i1][j1][k1];
    local deed2 = _DEED_LOG_PAGE_TAB_CONTENTS[i2][j2][k2];

    if (deed1.PAIRED == nil or
        deed2.PAIRED == nil) then
        return false;
    end

    -- A deed like Mines of Moria is paired with multiple class deeds
    -- Iterate through each deed's list of paired deeds and check for a match with these deeds.
    for deed1paired=1, table.maxn(deed1.PAIRED) do
        for deed2paired=1, table.maxn(deed2.PAIRED) do
            if ((deed1.PAIRED[deed1paired]["i"] == i2 and deed1.PAIRED[deed1paired]["j"] == j2 and deed1.PAIRED[deed1paired]["k"] == k2) and
                (deed2.PAIRED[deed2paired]["i"] == i1 and deed2.PAIRED[deed2paired]["j"] == j1 and deed2.PAIRED[deed2paired]["k"] == k1)) then
                return true;
            end
        end
    end

    return false;
end

function GetClassName(class)
    if (class == Turbine.Gameplay.Class.Beorning) then return "Beorning"; end
    if (class == Turbine.Gameplay.Class.Burglar) then return "Burglar"; end
    if (class == Turbine.Gameplay.Class.Captain) then return "Captain"; end
    if (class == Turbine.Gameplay.Class.Champion) then return "Champion"; end
    if (class == Turbine.Gameplay.Class.Guardian) then return "Guardian"; end
    if (class == Turbine.Gameplay.Class.Hunter) then return "Hunter"; end
    if (class == Turbine.Gameplay.Class.LoreMaster) then return "Lore Master"; end
    if (class == Turbine.Gameplay.Class.Minstrel) then return "Minstrel"; end
    if (class == Turbine.Gameplay.Class.RuneKeeper) then return "Rune-Keeper"; end
    if (class == Turbine.Gameplay.Class.Warden) then return "Warden"; end
    return "Unknown";
end

function GetVocationName(vocation)
    if (vocation == Turbine.Gameplay.Vocation.Armorer) then return "Armorer"; end
    if (vocation == Turbine.Gameplay.Vocation.Armsman) then return "Armsman"; end
    if (vocation == Turbine.Gameplay.Vocation.Explorer) then return "Explorer"; end
    if (vocation == Turbine.Gameplay.Vocation.Historian) then return "Historian"; end
    if (vocation == Turbine.Gameplay.Vocation.None) then return "None"; end
    if (vocation == Turbine.Gameplay.Vocation.Tinker) then return "Tinker"; end
    if (vocation == Turbine.Gameplay.Vocation.Woodsman) then return "Woodsman"; end
    if (vocation == Turbine.Gameplay.Vocation.Yeoman) then return "Yeoman"; end
    return "Unknown";
end

function CheckCharData()
    -- Deed Tracker doesn't support Session Play deeds
    if (IsSessionPlay()) then
        return;
    end

    -- This function checks if there is an entry in the _CHARDATA table, if not then it creates it.
    local CHARNAME = MYCHAR:GetName();

    if type (_CHARDATA[CHARNAME]) ~= 'table' then _CHARDATA[CHARNAME] = {} end;
    _CHARDATA[CHARNAME]["CLASS"] = MYCHAR:GetClass();
    _CHARDATA[CHARNAME]["VOCATION"] = MYCHAR:GetAttributes():GetVocation();
    _CHARDATA[CHARNAME]["RACE"] = MYCHAR:GetRace();
    _CHARDATA[CHARNAME]["LEVEL"] = MYCHAR:GetLevel();
    RegisterForLevelChange();
    CheckDeedData(CHARNAME);
end

function IsLoggedOnCharacter(character)
    return MYCHAR:GetName() == character;
end

function IsSessionPlay()
    local name = MYCHAR:GetName();
    return name:sub(1,1) == "~";
end

function GetLoggedOnCharacterName()
    local name = MYCHAR:GetName();
    if (name:sub(1,1) == "~") then
        name = name:sub(2);
    end
    return name;
end

function RegisterForLevelChange()
    MYCHAR.LevelChanged = function(sender, args)
        local level = sender:GetLevel();
        LevelChanged(level);
    end
end

function AnnounceNewSkirmishes(level)
    level = tostring(level);

    if (_LANG.SKIRMISHES.LEVELS[level] ~= nil) then
        local skirmishCount = _LANG.SKIRMISHES.LEVELS[level].COUNT;
        local skirmishes = GetString(_LANG.SKIRMISHES.LEVELS[level]);

        if (skirmishCount == 1) then
            Info(string.format(GetString(_LANG.SKIRMISHES.NEW_SKIRMISH), level, skirmishCount) .. skirmishes);
        elseif (skirmishCount > 1) then
            Info(string.format(GetString(_LANG.SKIRMISHES.NEW_SKIRMISHES), level, skirmishCount) .. skirmishes);
        end
    end
end

function LevelChanged(level)
    if (level == nil or level == "" or level == 0) then return; end

    local charName = MYCHAR:GetName();
    local mainWin = DeedTrackerWin.GetInstance();
    local uiCharacter = mainWin:GetUiCharacter();

    _CHARDATA[charName]["LEVEL"] = level;
    Debug("Deed Tracker thinks you changed to level " .. level);
    AnnounceNewSkirmishes(level);
    CheckDeedData(charName);

    if (charName == uiCharacter) then
        mainWin:LevelChanged();
    end
end

function MakeSaveEntry(isDeedComplete)
    local nowString = "";
    local method = "";
    if (isDeedComplete) then
        local now = Turbine.Engine.GetDate();
        nowString = string.format(
            "%04d-%02d-%02d %02d:%02d:%02d",
            now.Year, now.Month, now.Day, now.Hour, now.Minute, now.Second);
        if (IS_DEED_DETECTED) then 
            method = "Auto";
        elseif (IS_IMPORT_HAPPENING) then
            method = "Import";
        else
            method = "Manual"; 
        end
    end

    local saveData = {};

    saveData["COMPLETE"] = isDeedComplete;
    saveData["WHEN"] = nowString;
    saveData["METHOD"] = method;

    return saveData;
end

function IsCategory(deed)
    local isCategory = false;
    local deedType = GetString(_DEED_TYPES[deed.TYPE], "EN");
    if (deedType == "Category") then
        isCategory = true;
    end
    return isCategory;
end

function IsLegendaryDeed(deed)
    local isLegendaryDeed = false;

    if (deed.LEGENDARY ~= nil and deed.LEGENDARY) then
        isLegendaryDeed = true;
    end

    return isLegendaryDeed;
end

function IsDifficultyDeed(deed)
    local isDifficultyDeed = false;

    if (deed.DIFFICULTY ~= nil and deed.DIFFICULTY) then
        isDifficultyDeed = true;        
    end

    return isDifficultyDeed;
end

function GetDeedSkipInfo(character, currentDeed, i, j, k)
    local crv = GetCrv(currentDeed.CRV);
    local hasSubtype = DoesDeedHaveSubtype(currentDeed.SUBTYPE);
    local class = GetClass(character);
    local level = tonumber(GetLevel(character));
    local race = GetRace(character);
    local vocation = GetVocation(character);
    local isAlreadyComplete = GetDeedComplete(character, i, j, k);
    local isLegendaryDeed = IsLegendaryDeed(currentDeed);

    local isLegendaryServer = LoadServerField("LEGENDARY_SERVER");
    local isDifficultyServer = LoadServerField("DIFFICULTY_SERVER");

    local hideDeed = false;
    local skipDeed = false;
    local lvlTooHigh = false;
    local isCategory = false;
    local isNotAvailable = IsDeedNotAvailable(currentDeed.NA);
    local isDifficultyDeed = IsDifficultyDeed(currentDeed);

    if (crv == "Class" and not isAlreadyComplete) then
        -- if there is a sub-type, make sure it matches the display character:
        if (hasSubtype and currentDeed.SUBTYPE ~= class) then
            skipDeed = true;
        end
    elseif(crv == "Race" and not isAlreadyComplete) then
        if (hasSubtype and currentDeed.SUBTYPE ~= race) then
            skipDeed = true;
        end
    elseif(crv == "Vocation" and not isAlreadyComplete) then
        if (hasSubtype and currentDeed.SUBTYPE ~= vocation) then
            skipDeed = true;
        end
    end
    
    if(IsCategory(currentDeed)) then
        isCategory = true;
    end

    local minLevel = GetDeedMinimumLevel(currentDeed);
    if (SETTINGS.HIDE_DEEDS_ABOVE_LEVEL and
        level > 0 and
        minLevel > level) then
        lvlTooHigh = true;
    end

    if (SETTINGS.HIDE_DEEDS_BELOW_LEVEL and
        currentDeed.MAX_LVL ~= nil and 
        level > 0 and
        currentDeed.MAX_LVL < level) then
        skipDeed = true;
    end

    if (SETTINGS.HIDE_COMPLETED_DEEDS and
        isAlreadyComplete) then
        hideDeed = true;
    end

    if (isLegendaryDeed and not isLegendaryServer) then
        skipDeed = true;
    end

    if (isDifficultyDeed and not isDifficultyServer) then
        skipDeed = true;
    end

    if (SETTINGS.HIDE_NOT_ACTIVELY_ACHIEVABLE_DEEDS and 
        isNotAvailable and
        not isAlreadyComplete) then
        hideDeed = true;
    end

    return hideDeed, skipDeed, lvlTooHigh, isCategory, isNotAvailable;

end

function CheckDeedData(CHARACTER)

    -- Deed Tracker doesn't support Session Play deeds
    if (IsSessionPlay()) then
        return;
    end

    local _TABLE = _CHARDATA[CHARACTER];
    if type(_TABLE) ~= 'table' then return end;

    -- Pre-populate the completed state if it's not present.
    if _TABLE["DEEDS"] == nil then _TABLE["DEEDS"] = {} end;

    local totalCompleted = 0;
    totalDeedCount = 0;
    totalLpAvailable = 0;
    totalVxpAvailable = 0;

    for pageNumber=1, table.maxn(_DEED_LOG_PAGES) do
        if _TABLE["DEEDS"][pageNumber] == nil then _TABLE["DEEDS"][pageNumber] = {} end;
        for tabNumber=1, table.maxn(_DEED_LOG_PAGE_TABS[pageNumber]) do
            if _TABLE["DEEDS"][pageNumber][tabNumber] == nil then _TABLE["DEEDS"][pageNumber][tabNumber] = {} end;
            for deedNumber=1, table.maxn(_DEED_LOG_PAGE_TAB_CONTENTS[pageNumber][tabNumber]) do
                local currentDeed = _DEED_LOG_PAGE_TAB_CONTENTS[pageNumber][tabNumber][deedNumber];
                -- Get the save index for this deed:
                -- skip categories:
                if (not IsCategory(currentDeed)) then
                    local saveIndex = GetDeedSaveIndex(pageNumber, tabNumber, deedNumber);

                    if (_TABLE["DEEDS"][pageNumber][tabNumber][saveIndex] == nil) then
                        _TABLE["DEEDS"][pageNumber][tabNumber][saveIndex] = MakeSaveEntry(false);
                    elseif (type(_TABLE["DEEDS"][pageNumber][tabNumber][saveIndex])=="boolean") then
                        -- old system, convert to new:
                        _TABLE["DEEDS"][pageNumber][tabNumber][saveIndex] = 
                            MakeSaveEntry(_TABLE["DEEDS"][pageNumber][tabNumber][saveIndex]);
                    end

                    -- Count this deed if it applies to this character:
                    local hideDeed, skipDeed, lvlTooHigh, isCategory, isNotAvailable = 
                        GetDeedSkipInfo(CHARACTER, currentDeed, pageNumber, tabNumber, deedNumber);

                    if (not hideDeed and not skipDeed and not lvlTooHigh and not isNotAvailable) then
                        totalDeedCount = totalDeedCount + 1;

                        if (_TABLE["DEEDS"][pageNumber][tabNumber][saveIndex].COMPLETE) then
                            totalCompleted = totalCompleted + 1;
                        end

                        local completed = GetDeedComplete(CHARACTER, pageNumber, tabNumber, deedNumber);
                        if (not completed) then
                            totalLpAvailable = totalLpAvailable + currentDeed.LP;
                            totalVxpAvailable = totalVxpAvailable + currentDeed.VXP;
                        end
                    end
                end
            end
        end
    end

    _TABLE["DEEDS"]["COMPLETED"] = totalCompleted;

end

function GetStartingCharacterName()
    local name = MYCHAR:GetName();
    if (IsSessionPlay()) then
        return name:sub(2);
    else
        return name;
    end
end

function GetAltNames()
    local charName = GetStartingCharacterName();

    -- This function gets the names of the alts which have data stored 
    -- and builds the names into a returned table.
    local _ALTS = {};

    -- If we already have a list of characters, use that.
    if (_CHARDATA[":CHARACTERS"] ~= nil) then
        _ALTS = _CHARDATA[":CHARACTERS"];
    -- Otherwise, create one from the character data that is loaded:
    else
        for k,v in pairs (_CHARDATA) do
            if (k:sub(1, 1) ~= ":") then
                table.insert(_ALTS,k);
            end
        end
    end

    -- Check if the current character is there:
    local missingCurrentChar = true;
    for k,v in pairs (_ALTS) do
        if (v == charName) then
            missingCurrentChar = false;
        end
    end

    -- If the current character is missing...
    if (missingCurrentChar) then
        -- Add it to the list.
        table.insert(_ALTS, charName);

        -- It's missing because this character is new.
        -- See if there's any saved data to capture:
        loadCharData(charName);

        -- Go ahead and (re-)generate the physical file here:
        saveCurrentCharacterData();
    end

    table.sort(_ALTS);

    return _ALTS;
end

function GetDeedObjectiveControl(objective, width)
    local objectiveText = GetString(objective);
    if (objectiveText == nil or objectiveText == "") then
        return nil;
    end

    local deedObjectiveControl = Turbine.UI.Control();
    deedObjectiveControl:SetWidth(width);
    deedObjectiveControl:SetMouseVisible(false);

    local y = 0;
    local font = Turbine.UI.Lotro.Font.Verdana16;
    local debugColors = false;

    local objectiveLabel = Turbine.UI.Label();
    objectiveLabel:SetParent(deedObjectiveControl);
    objectiveLabel:SetFont(font);
    objectiveLabel:SetForeColor(Turbine.UI.Color.LightBlue);
    objectiveLabel:SetText(objectiveText);
    --objectiveLabel:SetText(GetString(_LANG.COMPLETE_WIN.WINDOW_DEED_PREFIX) .. deedName);
    objectiveLabel:SetSize(width - 50, 20);
    AutoFitLabelHeight(objectiveLabel, 1000);
    objectiveLabel:SetPosition(25, y);
    if (debugColors) then objectiveLabel:SetBackColor(Turbine.UI.Color.Blue); end
    objectiveLabel:SetMouseVisible(false);
    y = y + objectiveLabel:GetHeight();

    y = y + 5;
    deedObjectiveControl:SetHeight(y);
    return deedObjectiveControl;
end

function GetDeedGeoObjectiveControl(location, width)
    local deedGeoObjectiveControl = Turbine.UI.Control();
    deedGeoObjectiveControl:SetWidth(width);
    deedGeoObjectiveControl:SetMouseVisible(false);

    local pointCount = table.maxn(location.POINTS);

    local y = 0;
    local font = Turbine.UI.Lotro.Font.Verdana16;
    local debugColors = false;

    local objectiveOrNameText = GetString(location.PROGRESS);
    if (objectiveOrNameText == "") then
        objectiveOrNameText = GetString(location.NAME);
    end
    local waypointCommand = "";
    if (pointCount == 1 and SETTINGS.OBJECTIVES_SHOW_COORDINATES) then 
        local latLong = location.POINTS[1].LAT .. ", " .. location.POINTS[1].LONG;
        objectiveOrNameText = objectiveOrNameText .. " (" .. latLong .. ")";
        waypointCommand = "/way target " .. latLong;
    end

    if (WAYPOINT_AVAILABLE and waypointCommand ~= "") then
        -- Quickslot to trigger the Waypoint plugin:
        local quickslotButton = Turbine.UI.Button();
        quickslotButton:SetParent(deedGeoObjectiveControl);
        quickslotButton:SetSize(16,15);
        quickslotButton:SetPosition(5, y);
    
        local quickslot = Turbine.UI.Lotro.Quickslot();
        quickslot:SetParent(quickslotButton);
        quickslot:SetShortcut(Turbine.UI.Lotro.Shortcut(Turbine.UI.Lotro.ShortcutType.Alias, waypointCommand));
        quickslot:SetAllowDrop(false);
    
        quickslot.hider = Turbine.UI.Control();
        quickslot.hider:SetParent(quickslotButton);
        quickslot.hider:SetMouseVisible(false);
        quickslot.hider:SetBackground(_IMAGES.WAYPOINT_ARROW);
        quickslot.hider:SetBlendMode(Turbine.UI.BlendMode.AlphaBlend);
    end

    if (objectiveOrNameText) then
        local objectiveOrName = Turbine.UI.Label();
        objectiveOrName:SetParent(deedGeoObjectiveControl);
        objectiveOrName:SetFont(font);
        objectiveOrName:SetForeColor(Turbine.UI.Color.Yellow);
        objectiveOrName:SetText(objectiveOrNameText);
        --objectiveOrName:SetText(GetString(_LANG.COMPLETE_WIN.WINDOW_DEED_PREFIX) .. deedName);
        objectiveOrName:SetSize(width - 50, 20);
        AutoFitLabelHeight(objectiveOrName, 200);
        objectiveOrName:SetPosition(25, y);
        if (debugColors) then objectiveOrName:SetBackColor(Turbine.UI.Color.Blue); end
        objectiveOrName:SetMouseVisible(false);
        y = y + objectiveOrName:GetHeight();
    end

    local loreText = GetString(location.LORE);
    if (loreText) then
        local lore = Turbine.UI.Label();
        lore:SetParent(deedGeoObjectiveControl);
        lore:SetFont(Turbine.UI.Lotro.Font.Verdana12);
        lore:SetText(loreText);
        --lore:SetText(GetString(_LANG.COMPLETE_WIN.WINDOW_DEED_PREFIX) .. deedName);
        lore:SetSize(width - 75, 20);
        AutoFitLabelHeight(lore, 1000);
        lore:SetPosition(50, y);
        if (debugColors) then lore:SetBackColor(Turbine.UI.Color.Blue); end
        lore:SetMouseVisible(false);
        y = y + lore:GetHeight();
    end

    if (pointCount > 1) then
        -- add a label for each point:
        for i=1, pointCount do
            local point = location.POINTS[i];

            local pointLabel = Turbine.UI.Label();
            pointLabel:SetParent(deedGeoObjectiveControl);
            pointLabel:SetFont(Turbine.UI.Lotro.Font.Verdana12);
            pointLabel:SetForeColor(Turbine.UI.Color.LightYellow);
            pointLabel:SetText(i .. ": " .. point.LAT .. ", " .. point.LONG);
            pointLabel:SetSize(width - 75, 20);
            AutoFitLabelHeight(pointLabel, 200);
            pointLabel:SetPosition(50, y);
            if (debugColors) then pointLabel:SetBackColor(Turbine.UI.Color.Blue); end
            pointLabel:SetMouseVisible(false);
            y = y + pointLabel:GetHeight();
        end
    end

    y = y + 5;
    deedGeoObjectiveControl:SetHeight(y);
    return deedGeoObjectiveControl;
end

function GetIndiciesFromIJK(i, j, k)
    local saveIndex = GetDeedSaveIndex(i, j, k);
    return {i, j, k, saveIndex};
end

function GetIJKFromIndicies(indicies)
    local i, j, k, saveIndex = unpack(indicies);
    -- In the beginning, Completion Window data used K instead of SaveIndex.
    -- If SaveIndex is present, use it to override any saved K value.
    -- Otherwise, hope the K value is still valid.
    if (saveIndex ~= nil and saveIndex > 0) then
        k = GetCurrentKFromSaveIndex(i, j, saveIndex);
    end
    return {i, j, k};
end

function GetDeedInformationControl(i, j, k, width, includeObjectives)
    local y = 0;
    local font = Turbine.UI.Lotro.Font.Verdana16;
    local debugColors = false;
    local deedControl = Turbine.UI.Control();
    deedControl:SetWidth(width);

    local saveIndex = GetDeedSaveIndex(i, j, k);
    deedControl.Indicies = {i, j, k, saveIndex};

    deedControl:SetMouseVisible(false);

    if (debugColors) then deedControl:SetBackColor(Turbine.UI.Color.Yellow); end

    -- label for the deed name:
    local deedName = GetFullyQualifiedDeedName(i,j,k);
    local deedLabel = Turbine.UI.Label();
    deedLabel:SetParent(deedControl);
    deedLabel:SetFont(font);
    deedLabel:SetForeColor(Turbine.UI.Color.Yellow);
    deedLabel:SetText(GetString(_LANG.COMPLETE_WIN.WINDOW_DEED_PREFIX) .. deedName);
    deedLabel:SetSize(width - 50, 20);
    AutoFitLabelHeight(deedLabel, 200);
    deedLabel:SetPosition(10, y);
    if (debugColors) then deedLabel:SetBackColor(Turbine.UI.Color.Blue); end
    deedLabel:SetMouseVisible(false);

    y = y + deedLabel:GetHeight();

    -- label for the objective, if present:
    if (not includeObjectives or not SETTINGS.OBJECTIVES_SHOW_FULL_OBJECTIVES) then
        local summary = GetString(_DEED_LOG_PAGE_TAB_CONTENTS[i][j][k].SUMMARY);
        if (summary ~= "") then
            local summaryLabel = Turbine.UI.Label();
            summaryLabel:SetParent(deedControl);
            summaryLabel:SetFont(font);
            summaryLabel:SetMultiline(true);
            summaryLabel:SetText(GetString(_LANG.COMPLETE_WIN.WINDOW_DEED_SUMMARY) .. summary);
            summaryLabel:SetSize(width - 75, 20);
            AutoFitLabelHeight(summaryLabel, 200);
            summaryLabel:SetPosition(25, y);
            if (debugColors) then summaryLabel:SetBackColor(Turbine.UI.Color.Red); end
            summaryLabel:SetMouseVisible(false);

            y = y + summaryLabel:GetHeight() + 5;
        end
    end

    -- label for the lore, if present:
    local lore = GetString(_DEED_LOG_PAGE_TAB_CONTENTS[i][j][k].LORE);
    if (lore ~= "") then
        local loreLabel = Turbine.UI.Label();
        loreLabel:SetParent(deedControl);
        loreLabel:SetFont(font);
        loreLabel:SetMultiline(true);
        loreLabel:SetText(GetString(_LANG.COMPLETE_WIN.WINDOW_DEED_LORE) .. lore);
        loreLabel:SetSize(width - 75, 20);
        AutoFitLabelHeight(loreLabel, 1000);
        loreLabel:SetPosition(25, y);
        if (debugColors) then loreLabel:SetBackColor(Turbine.UI.Color.DarkGray); end
        loreLabel:SetMouseVisible(false);

        y = y + loreLabel:GetHeight() + 5;
    end

    -- label for the title, if present:
    local title = GetString(_DEED_LOG_PAGE_TAB_CONTENTS[i][j][k].TITLE);
    if (title ~= "") then
        local titleLabel = Turbine.UI.Label();
        titleLabel:SetParent(deedControl);
        titleLabel:SetFont(font);
        titleLabel:SetMultiline(true);
        titleLabel:SetText(GetString(_LANG.COMPLETE_WIN.WINDOW_DEED_TITLE) .. title);
        titleLabel:SetSize(width - 75, 20);
        AutoFitLabelHeight(titleLabel, 200);
        titleLabel:SetPosition(25, y);
        if (debugColors) then titleLabel:SetBackColor(Turbine.UI.Color.DarkBlue); end
        titleLabel:SetMouseVisible(false);

        y = y + titleLabel:GetHeight();
    end

    if (includeObjectives and SETTINGS.OBJECTIVES_SHOW_FULL_OBJECTIVES) then
        -- show the location of each objective if the data is available.
        local deed = _DEED_LOG_PAGE_TAB_CONTENTS[i][j][k];
        if (deed.ID ~= nil) then
            if (_OBJECTIVES[deed.ID] ~= nil) then
                local objectives = _OBJECTIVES[deed.ID];

                for i=1, table.maxn(objectives) do
                    -- there can be gaps, make sure it's there:
                    if (objectives[i] ~= nil) then
                        -- if it has geo stuff, make a geo control:
                        if (objectives[i].POINTS ~= nil) then
                            local deedGeoObjectiveControl = GetDeedGeoObjectiveControl(objectives[i], width);
                            if (deedGeoObjectiveControl ~= nil) then
                                deedGeoObjectiveControl:SetParent(deedControl);
                                deedGeoObjectiveControl:SetPosition(0, y);
                                y = y + deedGeoObjectiveControl:GetHeight();
                            end
                        -- otherwise, make a basic objective control:
                        else
                            local deedObjectiveControl = GetDeedObjectiveControl(objectives[i], width);
                            if (deedObjectiveControl ~= nil) then
                                deedObjectiveControl:SetParent(deedControl);
                                deedObjectiveControl:SetPosition(0, y);
                                y = y + deedObjectiveControl:GetHeight();
                            end
                        end
                    end
                end
            end
        end
    end

    deedControl:SetHeight(y);
    return deedControl;
end
